class EC3URLs:
    """
    This class is intended to store commonly used EC3 URLs.
    Should EC3 URLs change in the future they can be updated here
    """

    def __init__(self, response_format="json"):
        self.format = response_format

        self.base_url = "https://buildingtransparency.org/api/"

        # materials
        self.materials = "materials"
        self.material_statistics = "materials/statistics.{format}".format(
            format=self.format
        )
        self.material_statistics_cached = "materials/statistics/cached.{format}".format(
            format=self.format
        )
        self.material_epds_xpd_uuid = "materials?open_xpd_uuid={epd_xpd_uuid}".format(
            epd_xpd_uuid="{epd_xpd_uuid}"
        )

        # epds
        self.epds = "epds"
        self.epds_xpd_uuid = "epds/{epd_xpd_uuid}".format(
            format=self.format, epd_xpd_uuid="{epd_xpd_uuid}"
        )
        self.epds_id = "epds/{id}.{format}".format(format=self.format, id="{id}")

        # projects
        self.projects = "projects"
        self.projects_id = "projects?id={project_id}".format(
            format=self.format, project_id="{project_id}"
        )
        self.projects_name = "projects?name__like={project_name}".format(
            format=self.format, project_name="{project_name}"
        )

        # categories
        self.categories_tree = "categories/root"
        self.categories_id = "categories/{category_id}".format(
            format=self.format, category_id="{category_id}"
        )

    def base_url(self):
        """
        Returns the base url
        """
        return self.base_url

    ### MATERIALS ###
    def materials_url(self):
        """
        Combines the base URL and materials API URL
        """
        return self.base_url + self.materials

    def material_statistics_url(self):
        """
        Combines the base URL and materials statistics API URL
        """
        return self.base_url + self.material_statistics

    def material_statistics_cached_url(self):
        """
        Combines the base URL and materials statistics cached API URL
        """
        return self.base_url + self.material_statistics_cached

    def materials_xpd_uuid_url(self):
        """
        Combines the base URL and epds_xpd_uuid API URL
        """
        return self.base_url + self.material_epds_xpd_uuid

    ### EPDs ###
    def epds_url(self):
        """
        Combines the base URL and epds API URL
        """
        return self.base_url + self.epds

    def epds_xpd_uuid_url(self):
        """
        Combines the base URL and epds_xpd_uuid API URL
        """
        return self.base_url + self.epds_xpd_uuid

    def epds_id_url(self):
        """
        Combines the base URL and epds_id API URL
        """
        return self.base_url + self.epds_id

    def epds_product_classes_url(self):
        """
        Combines the base URL and epds_id API URL
        """
        return self.base_url + self.product_classes

    ### PROJECTS ###
    def projects_url(self):
        """
        Combines the base URL and projects API URL
        """
        return self.base_url + self.projects

    def projects_id_url(self):
        """
        Combines the base URL and projects ID API URL
        """
        return self.base_url + self.projects_id

    def projects_name_url(self):
        """
        Combines the base URL and projects ID API URL
        """
        return self.base_url + self.projects_name

    ### CATEGORIES ###
    def categories_tree_url(self):
        """
        Combines the base URL and categories url for the whole categories tree
        """
        return self.base_url + self.categories_tree

    def categories_id_url(self):
        """
        Combines the base URL and projects ID API URL
        """
        return self.base_url + self.categories_id



from .ec3_api import EC3Abstract
from .ec3_urls import EC3URLs


class EC3Projects(EC3Abstract):
    """
    Wraps functionality of EC3 Projects

    Usage:
        >>> ec3_project_list = EC3Projects(bearer_token=token, ssl_verify=False)
        >>> ec3_project_list.get_projects(params=project_param_dict)
    """

    def __init__(self, bearer_token, response_format="json", ssl_verify=True):
        super().__init__(
            bearer_token, response_format=response_format, ssl_verify=ssl_verify
        )

        self.sort_by = ""

        self.url = EC3URLs(response_format=response_format)

    def _process_params(self, params):
        params["params"]["page_size"] = self.page_size

        # NOTE "sort_by" is not currently working as expected when passing multiple fields.
        # Setting up to expect a single string field temporarily.
        if self.sort_by:
            params["params"]["sort_by"] = self.sort_by

        return params

    def get_projects(self, return_all=False, **params):
        """
        Returns matching Projects in your EC3 account

        Args:
            return_all (bool, optional): Set to True to return all matches. Defaults to False, which will return the quantity specified in page_size.

        Returns:
            list: List of dictionaries of matching Project records
        """
        processed_params = self._process_params(params)

        if return_all:
            return super()._get_all(self.url.projects_url(), **processed_params)
        else:
            return super()._get_records(self.url.projects_url(), **processed_params)

    def get_project_by_id(self, project_id):
        """
        Returns the project from a project id

        Args:
            project_id (str): Entity ID

        Returns:
            list: List of dictionaries of matching Project records
        """
        return super()._request(
            "get", self.url.projects_id_url().format(project_id=project_id)
        )

    def get_projects_by_name(self, project_name):
        """
        Returns a list of projects with a name equivalent to the input
        If your exact project name is put here you should get a list with one item.

        Args:
            project_name (str): Search term for your EC3 project name

        Returns:
            list: List of dictionaries of matching Project records
        """
        return super()._request(
            "get", self.url.projects_name_url().format(project_name=project_name)
        )



from datetime import datetime

from .ec3_api import EC3Abstract
from .ec3_urls import EC3URLs
from .ec3_categories import EC3Categories
from .ec3_utils import postal_to_latlong, get_masterformat_category_dict


class EC3Materials(EC3Abstract):
    """
    Wraps functionality of EC3 Materials

    Usage:
        >>> ec3_materials = EC3Materials(bearer_token=token, ssl_verify=False)
        >>> ec3_mat_list = ec3_materials.get_materials(params=mat_param_dict)
    """

    def __init__(self, bearer_token, response_format="json", ssl_verify=True):
        super().__init__(
            bearer_token, response_format=response_format, ssl_verify=ssl_verify
        )

        self.return_fields = []
        self.sort_by = ""
        self.only_valid = True
        self.masterformat_filter = (
            []
        )  # Currently EC3 requires you to go through category class for this

        self.url = EC3URLs(response_format=response_format)

    def _process_params(self, params):
        params["params"]["page_size"] = self.page_size

        if self.return_fields:
            fields_string = ",".join(self.return_fields)
            params["params"]["fields"] = fields_string

        # NOTE "sort_by" is not currently working as expected when passing multiple fields.
        # Setting up to expect a single string field temporarily.
        if self.sort_by:
            params["params"]["sort_by"] = self.sort_by

        if self.only_valid:
            params["params"]["epd__date_validity_ends__gt"] = datetime.today().strftime(
                "%Y-%m-%d"
            )

        if self.masterformat_filter:
            ec3_categories = EC3Categories(
                bearer_token=self.bearer_token, ssl_verify=False
            )
            whole_tree = ec3_categories.get_all_categories()
            masterformat_dict = get_masterformat_category_dict(whole_tree)

            category_ids = [masterformat_dict[i] for i in self.masterformat_filter]
            params["params"]["category"] = category_ids

        return params

    def get_materials(self, return_all=False, **params):
        """
        Returns matching materials

        Args:
            return_all (bool, optional): Set to True to return all matches. Defaults to False, which will return the quantity specified in max_records.

        Returns:
            list: List of dictionaries of matching material records
        """
        processed_params = self._process_params(params)

        if return_all:
            return super()._get_all(self.url.materials_url(), **processed_params)
        else:
            return super()._get_records(self.url.materials_url(), **processed_params)

    def get_materials_within_region(
        self,
        postal_code,
        country_code="US",
        plant_distance="100 mi",
        return_all=False,
        **params,
    ):
        """
        Returns only materials from plants within provided distance of postal code.
        This adds the "latitude", "longitude", and "plant_distance_lt" keys to your parameter dictionary.

        Args:
            postal_code (int): postal code
            country_code (str, optional): Two letter country code.. Defaults to 'US'.
            plant_distance (str, optional): Distance to plant with units in string ('mi' or 'km'). Defaults to "100 mi".
            return_all (bool, optional): Set to True to return all matches. Defaults to False, which will return the quantity specified in max_records.

        Returns:
            list: List of dictionaries of matching material records within distance provided from postal code
        """
        lat, long = postal_to_latlong(postal_code, country_code)
        params["params"]["latitude"] = lat
        params["params"]["longitude"] = long
        params["params"]["plant__distance__lt"] = plant_distance

        return self.get_materials(return_all=return_all, **params)

    # NOTE Querying materials by "open_xpd_uuid" does not appear to currently work with the api
    # def get_material_by_xpduuid(self, epd_xpd_uuid):
    #     """
    #     Returns the material from an Open xPD UUID of an EPD

    #     Args:
    #         epd_xpd_uuid (str): Open xPD UUID (Example: EC300001)

    #     Returns:
    #         list: List of dictionaries of matching material records
    #     """
    #     return super()._request(
    #         "get", self.url.materials_xpd_uuid_url().format(epd_xpd_uuid=epd_xpd_uuid)
    #     )



from .ec3_api import EC3Abstract
from .ec3_urls import EC3URLs


class EC3Categories(EC3Abstract):
    """
    Wraps functionality of EC3 Categories

    Usage:
        >>> ec3_categories = EC3Categories(bearer_token=token, ssl_verify=False)
        >>> ec3_categories.get_all_categories()
    """

    def __init__(self, bearer_token, response_format="json", ssl_verify=True):
        super().__init__(
            bearer_token, response_format=response_format, ssl_verify=ssl_verify
        )

        self.url = EC3URLs(response_format=response_format)

    def get_all_categories(self):
        """
        Gets the entire categories tree

        Returns:
            dict: Dictionary of entire category tree
        """
        return super()._request("get", self.url.categories_tree_url())

    def get_category_by_id(self, category_id):
        """
        Returns the category from a category id

        Args:
            category_id (str): Category ID

        Returns:
            dict: Returns a category by ID with the whole sub-categories tree
        """
        return super()._request(
            "get", self.url.categories_id_url().format(category_id=category_id)
        )



import abc
import requests


class EC3Abstract(metaclass=abc.ABCMeta):
    """
    Represents the abstract class for the Building Transparency Api

    This class get inherited by other major classes.
    """

    def __init__(self, bearer_token, response_format="json", ssl_verify=True):
        """
        Args:
            bearer_token (str): EC3 bearer token for the user
            response_format (str, optional): Defaults to "json".
            ssl_verify (bool, optional): Defaults to True.
        """

        session = requests.Session()
        self.bearer_token = bearer_token
        self.session = session
        self.session.headers.update({"Authorization": "Bearer {}".format(bearer_token)})

        self.page_size = 100  # Specifies the page size to return. Default is 100. Max allowed by api is 250
        self.max_records = 100  # Specifies the maximum number of records to return.

        self.format = response_format
        self._ssl_verify = ssl_verify
        if not ssl_verify:
            # ignore any unresolved references
            requests.packages.urllib3.disable_warnings()

        self.remove_nulls = True  # Determines if responses should return items with null values since EC3 returns everything

    def _process_response(self, response):
        """
        Processes the response. Removes null values if set to do so.

        Args:
            response (requests.models.Response): response from Api

        Returns:
            json: Processed records as json
        """
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            err_msg = str(exc)

            # Attempt to get Error message from response
            try:
                error_dict = response.json()
            except ValueError:
                pass
            else:
                if "error" in error_dict:
                    err_msg += " [Error: {}]".format(error_dict["error"])
            exc.args = (*exc.args, err_msg)
            raise exc
        else:
            # if user put in anything other than True or False, assume True
            if type(self.remove_nulls) != bool:
                self.remove_nulls = True

            if self.remove_nulls is True:
                ec3_response = response.json()
                if isinstance(ec3_response, dict):
                    cleaned_response = self._remove_nulls(ec3_response)
                else:
                    cleaned_response = [self._remove_nulls(d) for d in ec3_response]
                return cleaned_response
            else:
                return response.json()

    def _request(self, method, url, params=None):
        if params:
            response = self.session.request(
                method, url, verify=self._ssl_verify, params=params["params"]
            )
        else:
            response = self.session.request(method, url, verify=self._ssl_verify)

        return self._process_response(response)

    def _get_records(self, url, **params):
        """
        Returns the requested number of records.

        This will only return the first (or specified #) page when number of records exceeds page size.
        """
        data = self._request("get", url, params=params)
        requested_records = data
        page_number = 1

        while len(data) == self.page_size and len(requested_records) < self.max_records:
            page_number += 1
            params["params"]["page_number"] = page_number

            try:
                data = self._request("get", url, params=params)
            except requests.exceptions.HTTPError:
                break

            if data:
                requested_records.extend(data)

        if len(requested_records) > self.max_records:
            return requested_records[0 : self.max_records]
        else:
            return requested_records

    def _get_all(self, url, **params):
        """
        Returns all the records as a single list
        """
        hold_val1 = self.max_records
        self.max_records = 10000000  # temp max value

        data = self._get_records(url, **params)
        all_records = data
        page_number = 1

        while len(data) == self.page_size:
            page_number += 1
            params["params"]["page_number"] = page_number

            try:
                data = self._get_records(url, **params)
            except requests.exceptions.HTTPError:
                break

            if data:
                all_records.extend(data)

        self.max_records = hold_val1

        return all_records

    def _remove_nulls(self, response_dict):
        """
        Removes key/value pairs where value is None

        Args:
            response_dict (dict): Response dictionary

        Returns:
            dict: Cleaned version of input dictionary
        """
        for key, value in list(response_dict.items()):
            if value is None:
                del response_dict[key]
            elif isinstance(value, dict):
                self._remove_nulls(value)
        return response_dict



import pgeocode


def postal_to_latlong(postal_code, country_code="US"):
    """
    Converts postal code to latitude and longitude returned as array.
    Refer to pgeocode documentation for supported country codes.
    If not found, then coordinates for Null Island are returned (0,0).

    Args:
        postal_code (int): postal code
        country_code (str, optional): Two letter country code. Defaults to 'US'.

    Returns:
        A tuple containing a float (latitude) and a float (longitude)
    """
    lat = 0
    long = 0
    nomi = pgeocode.Nominatim(country_code)
    if nomi:
        df = nomi.query_postal_code(str(postal_code))
        lat = df["latitude"].item()
        long = df["longitude"].item()
    return (lat, long)


def recursive_dict_list_return(dict_item, key_name, out_keys, outlist=[]):
    """
    Recursively loops through a nested json/dictionary based on the key name

    This is intended for where the key_name may occur at multiple levels of nesting.
    For example, the "subcategories" key occurs at multiple levels of the EC3 categories tree.
    Using this function allows you to return a flattened list of dictionaries with the
    desired keys defined in the "out_keys" argument.

    Args:
        dict_item (dict): Dictionary with nested data to crawl through
        key_name (str): Name of key to crawl through nested dictionary
        out_keys (list[str]): List of key names to return
        outlist (list, optional): List to return (can be left empty if starting a new list)

    Returns:
        list: List of dictionaries with keys provided in out_keys
    """
    if (key_name in dict_item.keys()) and dict_item[key_name]:
        new_dict = {k: dict_item[k] for k in out_keys}
        outlist.append(new_dict)

        for d in dict_item[key_name]:
            recursive_dict_list_return(d, key_name, out_keys, outlist=outlist)

    elif key_name in dict_item:
        new_dict = {k: dict_item[k] for k in out_keys}
        outlist.append(new_dict)

    return outlist


def get_masterformat_category_dict(category_tree):
    """
    Get a dictionary with masterformat codes as the keys and ids as the values

    Args:
        category_tree (dict): This should be a nested dictionary of all or part of the category tree

    Returns:
        dict: Dictionary with masterformat codes as keys (ex: {'03 00 00 Concrete': '484df282d43f4b0e855fad6b351ce006'})
    """
    category_dict_list = recursive_dict_list_return(
        category_tree, "subcategories", ["masterformat", "id"]
    )
    masterformat_dict = {i["masterformat"]: i["id"] for i in category_dict_list}
    return masterformat_dict



__version__ = "0.0.8"

from .ec3_materials import EC3Materials
from .ec3_epds import EC3epds
from .ec3_urls import EC3URLs
from .ec3_projects import EC3Projects
from .ec3_categories import EC3Categories
